import botocore.response as btr

class ksu_bytes(bytes):
    ksu_metadata = {}

    @property
    def __class__(self):
        return bytes


class StreamingBody(btr.StreamingBody):
    ksu_metadata = {}

    def read(self, amt=None):
        result = super(StreamingBody, self).read(amt)
        if isinstance(result, bytes):
            result = ksu_bytes(result)
            result.ksu_metadata = self.ksu_metadata
        return result


import datetime
import getpass
import json
import logging
import os
import time
import jwt

from kensu.client import *
from kensu.utils.dsl.extractors.external_lineage_dtos import KensuDatasourceAndSchema
from kensu.utils.dsl import mapping_strategies
from kensu.utils.dsl.extractors import Extractors
from kensu.utils.dsl.lineage_builder import LineageBuilder
from kensu.utils.helpers import to_hash_key
from kensu.utils.injection import Injection
from kensu.pandas import DataFrame,Series
from kensu.utils.reporters import *
from kensu.utils.rule_engine import create_kensu_nrows_consistency
from kensu import sdk


class Kensu(object):
    UNKNOWN_PHYSICAL_LOCATION = PhysicalLocation(name="Unknown", lat=0.12341234, lon=0.12341234,
                                                 pk=PhysicalLocationPK(city="Unknown", country="Unknown"))

    @staticmethod
    def discover_user_name():
        return getpass.getuser()

    @staticmethod
    def get_git_repo():
        cur_dir = os.path.realpath(os.path.curdir)
        try:
            import git
            try:
                git_repo = git.Repo(cur_dir, search_parent_directories=True)
                return git_repo
            except git.GitError as e:
                logging.warning("kensu-py was unable to identify a git repo. The working dir is not a git repo?")
                pass
        except ImportError as e:
            logging.warning("Install GitPython for a maximum context about the GIT code repo if any")
            pass

    @staticmethod
    def discover_code_location():
        cur_dir = os.path.realpath(os.path.curdir)
        code_location = cur_dir
        git_repo = Kensu.get_git_repo()
        if git_repo is not None:
            remote = git_repo.remote()
            code_location = next(remote.urls, code_location)
        return code_location

    @staticmethod
    def discover_code_version():
        code_version = datetime.datetime.now().isoformat()

        git_repo = Kensu.get_git_repo()
        if git_repo is not None:
            code_version = git_repo.head.commit.hexsha
            if git_repo.is_dirty():
                code_version = code_version + " (dirty)"
        return code_version

    @staticmethod
    def get_conf_path(self, default = "conf.ini"):
        return os.environ["CONF_FILE"] if "CONF_FILE" in os.environ else default

    @staticmethod
    def build_conf():
        """
        Returns configparser.ConfigParser
        """
        from configparser import ConfigParser, ExtendedInterpolation
        config = ConfigParser(interpolation=ExtendedInterpolation())
        conf_path = Kensu.get_conf_path("conf.ini")
        try:
            config.read(conf_path)
        except:
            logging.warning(f"Cannot load config from file `%s`" % (conf_path))
        return config

    def __init__(self, api_url=None, auth_token=None, process_name=None,
                 user_name=None, code_location=None, init_context=True, 
                 do_report=None, report_to_file=None, offline_file_name=None, 
                 compute_stats=True, 
                 config=None, **kwargs):
        """
        config: : configparser.ConfigParser if None `build_conf` will be tried
        """
        if config is None:
            config = Kensu.build_conf()

        kensu_conf = config['kensu'] if config.has_section('kensu') else config['DEFAULT']
        self.conf = kensu_conf

        kensu_host = self.get_kensu_host(api_url)
        if kensu_host is None:
            kensu_host = kensu_conf.get("api_url")
        if auth_token is None:
            kensu_auth_token = os.environ.get('KENSU_API_TOKEN') or kensu_conf.get("api_token")
        else:
            kensu_auth_token = auth_token

        # returns a value following this precedence:
        #   kwargs = arg > conf > default
        # default is used to determine the type of the conf value (it can be overriden by tpe)
        def kwargs_or_conf_or_default(key, default, arg=None, kw=kwargs, conf=kensu_conf, tpe=None):
            if arg is not None:
                return arg
            elif key in kw and kw[key] is not None:
                return kw[key]
            elif key in conf and conf.get(key) is not None:
                if default is not None and tpe is None:
                    tpe = type(default)
                r = conf.get(key)
                if tpe is list:
                    r = r.split(",")
                elif tpe is bool:
                    r = conf.getboolean(key)
                elif tpe is not None:
                    r = tpe(r)
                return r
            else:
                return default
        self.extractors = Extractors()
        pandas_support = kwargs_or_conf_or_default("pandas_support", True)
        sklearn_support = kwargs_or_conf_or_default("sklearn_support", False)
        bigquery_support = kwargs_or_conf_or_default("bigquery_support", False)
        tensorflow_support = kwargs_or_conf_or_default("tensorflow_support", False)
        matplotlib_support = kwargs_or_conf_or_default("matplotlib_support", False)

        self.extractors.add_default_supports(pandas_support=pandas_support, sklearn_support=sklearn_support,bigquery_support=bigquery_support,tensorflow_support=tensorflow_support, matplotlib_support = matplotlib_support)

        bigquery_headers = kwargs_or_conf_or_default("bigquery_headers", None)
        if bigquery_headers:
            self.bigquery_headers = bigquery_headers

        project_names = kwargs_or_conf_or_default("project_names", [])
        environment = kwargs_or_conf_or_default("environment", None)
        timestamp = kwargs_or_conf_or_default("timestamp", None, tpe=int)
        logical_naming = kwargs_or_conf_or_default("logical_naming", None)
        mapping = kwargs_or_conf_or_default("mapping", None)
        report_in_mem = kwargs_or_conf_or_default("report_in_mem", False)
        sql_util_url = kwargs_or_conf_or_default("sql_util_url", None)

        if "get_code_version" in kwargs and kwargs["get_code_version"] is not None:
            get_code_version = kwargs["get_code_version"]
        else:
            get_code_version = Kensu.discover_code_version

        process_name = kwargs_or_conf_or_default("process_name", "Missing Application Name", process_name)
        user_name = kwargs_or_conf_or_default("user_name", "Missing User Name", user_name)
        code_location = kwargs_or_conf_or_default("code_location", None, code_location)

        reporter = kwargs_or_conf_or_default('reporter', None)
        do_report = kwargs_or_conf_or_default("do_report", True, do_report)
        report_to_file = kwargs_or_conf_or_default("report_to_file", False, report_to_file)
        offline_file_name = kwargs_or_conf_or_default("offline_file_name", None, offline_file_name)
        compute_stats = kwargs_or_conf_or_default("compute_stats", True, compute_stats)
        input_stats = kwargs_or_conf_or_default("input_stats", True)
        compute_delta = kwargs_or_conf_or_default("compute_delta",False)
        if compute_delta and not input_stats:
            logging.warning("delta nrows stats (compute_delta=True) will not work without setting input_stats=True")
        raise_on_check_failure = kwargs_or_conf_or_default("raise_on_check_failure", False)

        self.kensu_api = KensuEntitiesApi()
        self.kensu_api.api_client.host = kensu_host
        self.kensu_api.api_client.default_headers["X-Auth-Token"] = kensu_auth_token
        self.api_url = kwargs_or_conf_or_default("api_url", None)
        self.report_to_file = report_to_file

        sdk_pat = os.environ.get('KENSU_SDK_PAT') or kwargs_or_conf_or_default("PAT", None)
        sdk_url = os.environ.get('KENSU_SDK_URL') or kwargs_or_conf_or_default("sdk_url", None)
        sdk_verify_ssl = kwargs_or_conf_or_default("sdk_verify_ssl", True)
        if sdk_pat is None:
            self.sdk = sdk.DoNothingSDK()
        else:
            if sdk_url is None:
                sdk_url = self.api_url.replace('-api', '')                
            self.sdk = sdk.SDK(sdk_url, sdk_pat, verify_ssl=sdk_verify_ssl)


        # add function to Kensu entities
        injection = Injection()

        # reporter could be either an instance of Reporter already, or None, or a String (the class name of the reporter)
        if isinstance(reporter, str) or config.has_section('kensu.reporter'):
            reporter = Reporter.create(config['kensu.reporter'], reporter)
        elif reporter is not None and hasattr(reporter, '__call__'):
            reporter = GenericReporter(config, reporter.__call__)

        injection.set_reporter(reporter)
        injection.set_do_report(do_report, offline_file_name=offline_file_name, report_to_file=report_to_file)
        injection.set_kensu_api(self.kensu_api)

        self.logical_naming = logical_naming
        self.mapping = mapping
        self.report_in_mem = report_in_mem
        self.compute_stats = compute_stats
        self.compute_delta = compute_delta
        self.raise_on_check_failure = raise_on_check_failure
        self.offline_file_name = offline_file_name
        self.sql_util_url = sql_util_url
        self.input_stats = input_stats

        self.set_default_physical_location(Kensu.UNKNOWN_PHYSICAL_LOCATION)
        # can be updated using set_default_physical_location
        self.init_context(process_name=process_name, user_name=user_name, code_location=code_location,
                          get_code_version=get_code_version, project_names=project_names, environment=environment, timestamp=timestamp)

    # sets the api url using host if passed, otherwise gets KENSU_API_URL
    def get_kensu_host(self, host=None):
        if host is None:
                kensu_host = os.environ.get("KENSU_API_URL")
        else:
            kensu_host = host

        return kensu_host

    def register_schema_name(self, ds, schema):
        name = ds.name
        if "in-mem" in name and ds.format is not None:
            name = name + " of format=" + str(ds.format or '?')
        self.schema_name_by_guid[schema.to_guid()] = name
        try:
            self.logical_name_by_guid[schema.to_guid()] = ds.categories[0].split("::")[1]
        except:
            None
        return schema

    def to_schema_name(self, s_guid):
        return self.schema_name_by_guid.get(s_guid) or s_guid

    def to_schema_names(self, s_guids):
        return list(set([self.to_schema_name(s_guid) for s_guid in s_guids]))

    def init_context(self, process_name=None, user_name=None, code_location=None, get_code_version=None, project_names=None,environment=None,timestamp=None):
        # list of triples i, o, mapping strategy
        # i and o are either one or a list of triples (object, DS, SC)
        self.dependencies = []
        self.dependencies_mapping = []
        self.dependencies_per_columns = {}
        # when set, it seems to indicate if a schema/DS is a real one (is not in memory),
        # so must be not set for in memory DSes
        self.real_schema_df = {}
        self.schema_name_by_guid = {}
        self.logical_name_by_guid = {}
        # by lineage_id, thus this tracks for which lineages a lineage run was already sent
        self.sent_runs = []
        self.data_collectors = {}
        self.model={}
        self.set_timestamp(timestamp)
        self.inputs_ds = []
        self.lineage_and_ds = {}
        self.write_reinit = False
        self.rules = []
        self.check_rules = []
        self.schema_stats = {}
        self.stats_to_send={}

        if user_name is None:
            user_name = Kensu.discover_user_name()
        if code_location is None:
            code_location = Kensu.discover_code_location()
        self.user = User(pk=UserPK(user_name))._report()
        self.code_base = CodeBase(pk=CodeBasePK(code_location))._report()
        if get_code_version is None:
            if timestamp is not None: # this is weird though...
                version = datetime.datetime.fromtimestamp(timestamp/1000).isoformat()
            else:
                version = Kensu.discover_code_version()
        else:
            version = get_code_version()
        self.code_version = CodeVersion(maintainers_refs=[self.user.to_ref()],
                                        pk=CodeVersionPK(version=version,
                                                         codebase_ref=self.code_base.to_ref()))._report()
        if process_name is None:
            if "__file__" in globals():
                process_name = os.path.basename(os.path.realpath(__file__))
            else:
                raise Exception("Can't determine `process_name`, maybe is this running from a Notebook?")
        self.process = Process(pk=ProcessPK(qualified_name=process_name))._report()
        if project_names is None:
            self.project_refs = []
        else:
            self.project_refs = [Project(pk=ProjectPK(name=n))._report().to_ref() for n in project_names]
        process_run_name = process_name + "@" + datetime.datetime.now().isoformat()
        self.process_run = ProcessRun(
            pk=ProcessRunPK(process_ref=self.process.to_ref(), qualified_name=process_run_name)
            , launched_by_user_ref=self.user.to_ref()
            , executed_code_version_ref=self.code_version.to_ref()
            , projects_refs=self.project_refs
            , environment = environment
        )._report()

    def set_reinit(self, bool = True):
        self.write_reinit = bool

    def add_input_ref(self, entities):
        if self.write_reinit == True:
            self.inputs_ds = []
            self.write_reinit = False
        self.inputs_ds.append(entities)

    def set_timestamp(self, timestamp):
        if timestamp is not None:
            self.kensu_api.api_client.default_headers["X-Entity-Creation-Time"] = timestamp
        else:
            timestamp = datetime.datetime.now().timestamp()*1000
        self.timestamp = timestamp

    def set_default_physical_location(self, pl):
        self.default_physical_location = pl
        pl._report()
        self.default_physical_location_ref = pl.to_ref()

    def get_dependencies_mapping(self):
        return self.dependencies_mapping

    def add_dependencies_mapping(self, guid, col, from_guid, from_col, type):
        dep = {'GUID': guid,
               'COLUMNS': col,
               'FROM_ID': from_guid,
               'FROM_COLUMNS': from_col,
               'TYPE': type}
        self.dependencies_mapping.append(dep)

    def in_mem(self, var_name):
        return "in-memory-data://" + self.process.pk.qualified_name + "/" + var_name

    def name_for_stats(self, input_name):
        return input_name.replace('.', '_')

    def extract_stats(self, stats_df):
        try:
            stats = self.extractors.extract_stats(stats_df)
        except Exception as e:
            logging.warning(f'stats extraction from {type(stats_df)} failed', e)
            if isinstance(stats_df, dict):
                # FIXME: this is weird logic here... we'd post the actual data instead of stats!
                stats = stats_df
            else:
                # TODO Support ndarray
                stats = None
        return stats


    def report_with_mapping(self):
        self.set_reinit()
        import pandas as pd
        deps = self.dependencies_mapping
        ddf = pd.DataFrame(deps)
        df = ddf.set_index(['GUID', 'COLUMNS', 'FROM_ID']).groupby(['GUID', 'COLUMNS', 'FROM_ID']).agg(list)
        df = df[~df.index.duplicated(keep='first')].reset_index()

        unique_ids = list(df['GUID'].unique())

        if self.report_in_mem:
            #TODO Deprecate
            for element in unique_ids:
                dataflow = []
                dependencies = df[df['GUID'] == element]
                data = {}
                for row in dependencies.iterrows():
                    info = row[1]
                    from_columns = [str(x) for x in info['FROM_COLUMNS']]
                    data[str(info['COLUMNS'])] = from_columns
                schema_dep = SchemaLineageDependencyDef(from_schema_ref=SchemaRef(by_guid=info['FROM_ID']),
                                                        to_schema_ref=SchemaRef(by_guid=info['GUID']),
                                                        column_data_dependencies=data)
                dataflow.append(schema_dep)

                lineage = ProcessLineage(name=self.get_lineage_name(dataflow),
                                         operation_logic='APPEND',
                                         pk=ProcessLineagePK(process_ref=ProcessRef(by_guid=self.process.to_guid()),
                                                             data_flow=dataflow))._report()

                if lineage.to_guid() not in self.sent_runs:
                    lineage_run = LineageRun(pk=LineageRunPK(lineage_ref=ProcessLineageRef(by_guid=lineage.to_guid()),
                                                             process_run_ref=ProcessRunRef(
                                                                 by_guid=self.process_run.to_guid()),
                                                             timestamp=round(self.timestamp)))._report()
                    self.sent_runs.append(lineage.to_guid())
        else:
            dependencies_per_columns = {}
            for element in unique_ids:
                if element in self.real_schema_df:
                    sub_df = df[df['GUID'] == element]
                    for row in sub_df.iterrows():
                        info = row[1]
                        destination_guid = info['GUID']
                        guid = info['GUID']
                        origin_column = info['COLUMNS']
                        column = info['COLUMNS']
                        all_deps = df
                        self.create_dependencies(destination_guid, guid, origin_column, column, all_deps,
                                            dependencies_per_columns)

            dataflows = {}
            for destination_guid in dependencies_per_columns:
                if dependencies_per_columns[destination_guid] != {}:
                    dataflows[destination_guid] = {}
                    for column in dependencies_per_columns[destination_guid]:
                        for origin_guid in dependencies_per_columns[destination_guid][column]:
                            if origin_guid not in dataflows[destination_guid] and origin_guid!=destination_guid:
                                dataflows[destination_guid][origin_guid] = {}
                                dataflows[destination_guid][origin_guid][column]=list(set(dependencies_per_columns[destination_guid][column][origin_guid]))
                            elif origin_guid!=destination_guid:
                                dataflows[destination_guid][origin_guid][column] = \
                                list(set(dependencies_per_columns[destination_guid][column][origin_guid]))

            for to_guid in dataflows:
                schemas_pk = set()
                from_pks = set()
                dataflow = []
                is_ml_model = False
                for from_guid in dataflows[to_guid]:

                    schema_dep = SchemaLineageDependencyDef(from_schema_ref=SchemaRef(by_guid=from_guid),
                                                            to_schema_ref=SchemaRef(by_guid=to_guid),
                                                            column_data_dependencies=dataflows[to_guid][from_guid])
                    if to_guid in self.model:
                        is_ml_model = True
                    dataflow.append(schema_dep)
                    schemas_pk.add(from_guid)
                    from_pks.add(from_guid)
                    schemas_pk.add(to_guid)


                lineage = ProcessLineage(name=self.get_lineage_name(dataflow),
                                         operation_logic='APPEND',
                                         pk=ProcessLineagePK(
                                             process_ref=ProcessRef(by_guid=self.process.to_guid()),
                                             data_flow=dataflow))._report()
                if lineage.to_guid() not in self.sent_runs:
                    from_schema_ref = set(x.from_schema_ref.by_guid for x in lineage.pk.data_flow)
                    to_schema_ref =  set(x.to_schema_ref.by_guid for x in lineage.pk.data_flow)
                    self.lineage_and_ds[lineage.to_guid()] = {'from_schema_ref':list(from_schema_ref),
                                                              'to_schema_ref':to_schema_ref}


                    lineage_run = LineageRun(
                        pk=LineageRunPK(lineage_ref=ProcessLineageRef(by_guid=lineage.to_guid()),
                                        process_run_ref=ProcessRunRef(by_guid=self.process_run.to_guid()),
                                        timestamp=round(self.timestamp)))._report()
                    self.sent_runs.append(lineage.to_guid())


                    def create_stats(schema_guid):
                        stats_df = self.real_schema_df[schema_guid]
                        stats = self.extract_stats(stats_df)
                        if stats is None and isinstance(stats_df, KensuDatasourceAndSchema):
                            # in this special case, the stats are reported not by kensu-py,
                            # but by some external library called from .f_publish_stats callback
                            # as an example - stats publishing directly from Apache Spark JVM
                            # for pyspark's python-JVM interop jobs (where data moves between Spark & Python)
                            stats_df.f_publish_stats(lineage_run.to_guid())
                        elif stats is not None:
                            self.schema_stats[schema_guid] = stats
                            if lineage_run.to_guid() not in self.stats_to_send:
                                self.stats_to_send[lineage_run.to_guid()] = {}
                            self.stats_to_send[lineage_run.to_guid()][schema_guid] = stats
                            if schema_guid == to_guid:
                                if 'nrows' in stats and self.compute_delta:
                                    # fixme: extract a helper
                                    output_nrows = stats['nrows']
                                    #TODO Corner case: if from_pk = to_pk
                                    for input_schema_guid in from_pks:
                                        input_stats = self.schema_stats.get(input_schema_guid)
                                        input_nrows = input_stats and input_stats.get('nrows')
                                        if input_nrows:
                                            delta_nrows = input_nrows - output_nrows
                                            input_name = self.logical_name_by_guid.get(input_schema_guid, self.schema_name_by_guid.get(input_schema_guid))
                                            clean_input_name = self.name_for_stats(input_name)
                                            stats['delta.nrows_' + clean_input_name + '.abs']=delta_nrows
                                            stats['delta.nrows_' + clean_input_name + '.decrease in %'] = round(100*delta_nrows/input_nrows,2)
                                    self.schema_stats[schema_guid] = stats
                                    self.stats_to_send[lineage_run.to_guid()][schema_guid] = stats


                        #FIXME should be using extractors instead
                        if is_ml_model:
                            model_name = self.model[to_guid][1]
                            metrics = self.model[to_guid][2]
                            import json
                            hyperparams = json.dumps(self.model[to_guid][3])

                            model = Model(ModelPK(name=model_name))._report()
                            train = ModelTraining(pk=ModelTrainingPK(model_ref=ModelRef(by_guid=model.to_guid()),
                                                                     process_lineage_ref=ProcessLineageRef(
                                                                         by_guid=lineage.to_guid())))._report()
                            r=ModelMetrics(pk=ModelMetricsPK(model_training_ref=ModelTrainingRef(by_guid=train.to_guid()),
                                                           lineage_run_ref=LineageRunRef(by_guid=lineage_run.to_guid()),
                                                           stored_in_schema_ref=SchemaRef(by_guid=to_guid)),
                                         metrics=metrics, hyper_params_as_json=hyperparams)._report()
                    for schema in from_pks:
                        if self.input_stats:
                            create_stats(schema)
                    create_stats(to_guid)


                    for check in self.check_rules:
                        if list(check.keys())[0] == 'nrows_consistency':
                            create_kensu_nrows_consistency(check[list(check.keys())[0]])
                    self.send_rules()

                    lineage_run_id = lineage_run.to_guid()
                    for lineage_run_id in self.stats_to_send:
                        for schema_id in self.stats_to_send[lineage_run_id]:
                            self.send_stats(lineage_run_id,schema_id,self.stats_to_send[lineage_run_id][schema_id])
                    if lineage_run_id in self.stats_to_send:
                        del self.stats_to_send[lineage_run_id]

                    self.check_local_rules()

    def create_dependencies(self,destination_guid, guid, origin_column, column, all_deps,
                            dependencies_per_columns_rt):
        visited = list()
        visited.append((guid,column))
        self.dependencies_per_columns = dependencies_per_columns_rt
        filtered_dependencies = all_deps[all_deps['GUID'] == guid]

        filtered_dependencies = filtered_dependencies[filtered_dependencies['COLUMNS'] == str(column)]
        if destination_guid in self.dependencies_per_columns:
            for row in filtered_dependencies.iterrows():
                row = row[1]

                if row['FROM_ID'] in self.real_schema_df:
                    if origin_column in self.dependencies_per_columns[destination_guid]:
                        if row['FROM_ID'] in self.dependencies_per_columns[destination_guid][origin_column]:
                            self.dependencies_per_columns[destination_guid][origin_column][row['FROM_ID']] = \
                            self.dependencies_per_columns[destination_guid][origin_column][row['FROM_ID']] + \
                            row['FROM_COLUMNS']
                        else:
                            self.dependencies_per_columns[destination_guid][origin_column][row['FROM_ID']] = row[
                                'FROM_COLUMNS']
                    else:
                        self.dependencies_per_columns[destination_guid][origin_column] = {}
                        self.dependencies_per_columns[destination_guid][origin_column][row['FROM_ID']] = row[
                            'FROM_COLUMNS']
                        # dependencies_per_columns[guid][row['FROM_ID']] = row['FROM_COLUMNS']
                else:
                    guid = row['FROM_ID']
                    columns = row['FROM_COLUMNS']
                    for column in columns:
                        if (guid,column) not in visited:
                            self.create_dependencies(destination_guid, guid, origin_column, column, all_deps,
                                                self.dependencies_per_columns)
        else:
            self.dependencies_per_columns[destination_guid] = {}
            self.create_dependencies(destination_guid, guid, origin_column, column, all_deps,
                                self.dependencies_per_columns)

    def get_dependencies(self):
        return self.dependencies
    def add_dependency(self, i, o, mapping_strategy=mapping_strategies.FULL):
        if not isinstance(i, tuple):
            (ids, isc) = self.extractors.extract_data_source_and_schema(i, self.default_physical_location_ref)
            i = (i, ids, isc)

        if not isinstance(o, tuple):
            (ods, osc) = self.extractors.extract_data_source_and_schema(o, self.default_physical_location_ref)
            o = (o, ods, osc)

        self.dependencies.append((i, o, mapping_strategy))
    def add_dependencies(self, ins, outs, mapping_strategy=mapping_strategies.FULL):
        new_ins = []
        for i in ins:
            if not isinstance(i, tuple):
                (ids, isc) = self.extractors.extract_data_source_and_schema(i, self.default_physical_location_ref)
                i = (i, ids, isc)
                new_ins.append(i)

        new_outs = []
        for o in outs:
            if not isinstance(o, tuple):
                (ods, osc) = self.extractors.extract_data_source_and_schema(o, self.default_physical_location_ref)
                o = (o, ods, osc)
                new_outs.append(o)

        self.dependencies.append((new_ins, new_outs, mapping_strategy))

    def get_lineage_name(self,
                         data_flow  # type: list[SchemaLineageDependencyDef]
                         ):
        inputs = ",".join(sorted(self.to_schema_names([d.from_schema_ref.by_guid for d in data_flow])))
        outputs = ",".join(sorted(self.to_schema_names([d.to_schema_ref.by_guid for d in data_flow])))
        return "Lineage to {} from {}".format(outputs, inputs)


    @property
    def s(self):
        return self.start_lineage(True)


    def start_lineage(self, report_stats=True):
        lineage_builder = LineageBuilder(self, report_stats)
        return lineage_builder


    def new_lineage(self, process_lineage_dependencies, report_stats=True, **kwargs):
        # if the new_lineage has a model training in it (output),
        #   then kwargs will be pass to the function to compute metrics
        #     ex: kwargs["y_test"] can refer to the test set to compute CV metrics
        data_flow = [d.toSchemaLineageDependencyDef() for d in process_lineage_dependencies]
        lineage = ProcessLineage(name=self.get_lineage_name(data_flow),
                                 operation_logic="APPEND",
                                 # FIXME? => add control and the function level like report_stats
                                 pk=ProcessLineagePK(
                                     process_ref=self.process.to_ref(),
                                     data_flow=data_flow
                                 )
                                 )._report()

        if self.timestamp is None:
            self.timestamp=int(time.time()) * 1000

        lineage_run = LineageRun(pk=LineageRunPK(
            lineage_ref=lineage.to_ref(),
            process_run_ref=self.process_run.to_ref(),
            timestamp=self.timestamp
        )
        )._report()

        data_flow_inputs = list(
            {to_hash_key(d.input_schema): (d.input_schema, d.input) for d in process_lineage_dependencies}.values())
        data_flow_outputs = list(
            {to_hash_key(d.output_schema): (d.output_schema, d.output) for d in process_lineage_dependencies}.values())

        for (schema, df) in (data_flow_inputs + data_flow_outputs):
            stats = self.extractors.extract_stats(df)
            if report_stats and stats is not None:
                DataStats(pk=DataStatsPK(schema_ref=schema.to_ref(),
                                         lineage_run_ref=lineage_run.to_ref()),
                          stats=stats,
                          extra_as_json=None)._report()

        # TODO Machine Learning part for OUTPUTS ONLY (right ?)
        for (schema, df) in data_flow_outputs:
            ml = self.extractors.extract_machine_learning_info(df)
            if ml is not None:
                model = Model(ModelPK(ml["name"]))._report()
                model_training = ModelTraining(ModelTrainingPK(model_ref=model.to_ref(),
                                                               process_lineage_ref=lineage.to_ref())
                                               )._report()

                metrics = self.extractors.extract_machine_learning_metrics(df, **kwargs)
                if len(metrics) > 0:
                    hp = self.extractors.extract_machine_learning_hyper_parameters(df)
                    ModelMetrics(pk=ModelMetricsPK(
                        model_training_ref=model_training.to_ref(),
                        lineage_run_ref=lineage_run.to_ref(),
                        stored_in_schema_ref=schema.to_ref()
                    ),
                        metrics=metrics,
                        hyper_params_as_json=json.dumps(hp)
                    )._report()

    def send_stats(self,lineage_run_id, schema_id,stats):

        if stats is not None:
            DataStats(pk=DataStatsPK(schema_ref=SchemaRef(by_guid=schema_id),
                                     lineage_run_ref=LineageRunRef(by_guid=lineage_run_id)),
                      stats=stats,
                      extra_as_json=None)._report()

    def send_rules(self):
        if self.rules and self.api_url:
            process_id = self.process.to_guid()
            project_id = self.process_run.projects_refs[0].by_guid
            env_name = self.process_run.environment
            cv = self.process_run.executed_code_version_ref.by_guid

            for map in self.rules:
                for lds_id in map:
                    field_name = map[lds_id]['field']
                    fun = map[lds_id]['fun']
                    context = map[lds_id]['context']

                    token_content = jwt.decode(self.kensu_api.api_client.default_headers["X-Auth-Token"], options={"verify_signature": False})
                    if 'SANDBOX_USER_HASH' in token_content:
                        sandbox_prefix = token_content['SANDBOX_USER_HASH'] + "-"
                        project_id = sandbox_prefix + project_id
                        process_id = sandbox_prefix + process_id
                        cv = sandbox_prefix +cv
                    else:
                        sandbox_prefix = ''

                    data = self.sdk.get_lineages_in_project(project_id, process_id, env_name, cv)
                    try:
                        lds_guid = [e['datasource'] for e in data['data']['nodes'] if e['datasource']['name'] == lds_id][0][
                            'id']
                    except:
                        try:
                            from kensu.utils.kensu_class_handlers import KensuClassHandlers
                            from kensu.client.models.logical_data_source import LogicalDataSourcePK
                            lds_guid = sandbox_prefix + KensuClassHandlers.guid_pk(LogicalDataSourcePK(name=lds_id,location=lds_id))
                        except:
                            lds_guid = None
                            logging.info("LDS Guid %s not in this lineage"%lds_id)

                    if lds_guid is not None:
                        if context == "DATA_STATS":
                            lineage_ids = set(
                                [data['data']['links'][i]['lineage']['id'] for i in range(len(data['data']['links']))])
                            lineage_id_for_ds={}
                            for link in data['data']['links']:
                                for lineage_id in lineage_ids:
                                    if link['lineage']['id'] == lineage_id:
                                        lineage_id_for_ds.setdefault(lineage_id,[])
                                        lineage_id_for_ds[lineage_id] = set(list(lineage_id_for_ds[lineage_id]) + [link['sourceDatasourceId']] + [link['targetDatasourceId']])

                            for lineage_id in lineage_id_for_ds:
                                if lds_guid in lineage_id_for_ds[lineage_id]:
                                    current_rules = self.sdk.get_rules_for_ds_in_project(lds_guid, lineage_id, project_id,
                                                                                         env_name)
                                    current_range_rules = {i['fieldName']: i['uuid'] for i in
                                                           current_rules['data']['predicates'] if
                                                           i['functionName'] == 'Range' and (i['environment'] == env_name)}
                                    current_frequency_rule = [i['uuid'] for i in current_rules['data']['predicates'] if
                                                              i['functionName'] == 'Frequency' and (
                                                                          i['environment'] == env_name)]
                                    if fun['name'] == 'Range' and (field_name in current_range_rules):
                                        self.sdk.update_rule(current_range_rules[field_name], fun)
                                    elif fun['name'] == 'Frequency' and current_frequency_rule:
                                        self.sdk.update_rule(current_frequency_rule[0], fun)

                                    else:
                                        self.sdk.create_rule(lds_guid, lineage_id, project_id, process_id, env_name, field_name,
                                                             fun)

                        elif context == "LOGICAL_DATA_SOURCE":
                            current_rules = self.sdk.get_all_rules_for_ds(lds_guid)

                            current_range_rules = {i['fieldName']: i['uuid'] for i in
                                                   current_rules['data']['predicates'] if
                                                   i['functionName'] == 'Range'}
                            current_variation_rules = {i['fieldName']: i['uuid'] for i in
                                                   current_rules['data']['predicates'] if
                                                   i['functionName'] == 'Variability'}
                            if fun['name'] == 'Range' and (field_name in current_range_rules):
                                self.sdk.update_rule(current_range_rules[field_name], fun)
                            elif fun['name'] == 'Variability' and (field_name in current_variation_rules):
                                self.sdk.update_rule(current_variation_rules[field_name], fun)

                            else:
                                self.sdk.create_rule(lds_id=lds_guid, field_name=field_name,
                                                     fun=fun,context = "LOGICAL_DATA_SOURCE")




    def check_local_rules(self):
        from kensu.utils.exceptions import NrowsConsistencyError
        for check in self.check_rules:
            if 'check_nrows_consistency' in check:
                for datasource in check['check_nrows_consistency']:
                    for rule in check['check_nrows_consistency'][datasource]:
                        for field in rule:
                            if rule[field]['output_nrows'] != rule[field]['input_nrows'] :
                                e = NrowsConsistencyError(datasource, rule[field]['input_nrows'],
                                                                rule[field]['output_nrows'])
                                if self.raise_on_check_failure:
                                    raise e
                                else:
                                    logging.warning(e)






import boto3
from boto3 import *

from kensu.client import DataSourcePK, DataSource, FieldDef, SchemaPK, Schema
from kensu.requests.models import ksu_str
from kensu.utils.kensu_provider import KensuProvider
from kensu.utils.helpers import logical_naming_batch

class ksu_dict(dict):
    ksu_metadata = {}

    @property
    def __class__(self):
        return dict


def kensu_put(event_params, event_ctx, **kwargs):
    if isinstance(event_params.get('Body'), ksu_str):
        kensu = KensuProvider().instance()
        put_body = event_params.get('Body')

        #The input is the ksu_str, the metadata contains its schema pk
        input_schema = put_body.metadata['real_schema']
        short_schema = put_body.metadata['short_schema']

        input_location = put_body.metadata['ds_location']
        s3_bucket = event_params.get('Bucket') or 'unknown-s3-bucket'
        s3_key = event_params.get('Key') or 'unknown-s3-key'

        # Creation of the output datasource (stored in S3)
        location = 'aws::S3::' + s3_bucket + '/' + s3_key
        name = s3_key

        if kensu.logical_naming == 'ReplaceNumbers':
            logical = logical_naming_batch(name)
        else:
            logical = name

        result_pk = DataSourcePK(location=location,
                                 physical_location_ref=kensu.default_physical_location_ref)
        result_ds = DataSource(name=name, categories=['logical::'+logical],format=name.split('.')[-1],
                               pk=result_pk)._report()

        input_fields = [k.name for k in short_schema.pk.fields]
        input_schema_pk = short_schema.to_guid()

        # This data source has the same schema fields as the input

        fields = [FieldDef(name=k.name, field_type=k.field_type, nullable=True) for k in input_schema.pk.fields]

        sc_pk = SchemaPK(result_ds.to_ref(),
                         fields=fields)
        result_sc = Schema(name="schema:" + result_ds.name, pk=sc_pk)._report()

        fields = [FieldDef(name=k.name, field_type=k.field_type, nullable=True) for k in short_schema.pk.fields]

        sc_pk = SchemaPK(result_ds.to_ref(),
                         fields=fields)

        short_result_sc = Schema(name="short-schema:" + result_ds.name, pk=sc_pk)._report()

        kensu.real_schema_df[short_result_sc.to_guid()] = put_body.metadata['stats']

        for col in input_fields:
            kensu.add_dependencies_mapping(short_result_sc.to_guid(),str(col),input_schema_pk,str(col),'s3_put')
        kensu.report_with_mapping()



def add_custom_method(class_attributes, **kwargs):
    class_attributes['kensu_put'] = kensu_put

    original_get = class_attributes['get']

    def kensu_get(event_params, **kwargs):
        result = original_get(event_params,**kwargs)

        if 'Body' in result:
            result = ksu_dict(result)

            s3_bucket = event_params.bucket_name or 'unknown-s3-bucket'
            s3_key = event_params.key or 'unknown-s3-key'

            # Creation of the output datasource (stored in S3)
            location = 'aws::S3::' + s3_bucket + '/' + s3_key
            name = s3_key
            result.ksu_metadata['origin_location'] = location
            result.ksu_metadata['origin_name'] = name

            import botocore.response as btr
            from kensu.botocore.response import StreamingBody

            if isinstance(result['Body'],btr.StreamingBody):
                result['Body'].__class__ = StreamingBody
                result['Body'].ksu_metadata = result.ksu_metadata
        return result

    class_attributes['get'] = kensu_get

boto3._get_default_session().events.register("creating-resource-class.s3.Object",
                        add_custom_method)



def kensu_tracker(*class_attributes, **kwargs):
    import logging
    param_types = [
    ]
    import pprint
    if kwargs.get('params'):
        param_types = [[k, v, type(v)] for k, v in kwargs.get('params').items()]
    logging.debug('---\nKensu AWS tracker: '
          'param_types: {}\n'
          'class_attributes:{}\n kwargs: {}\n-----'.format(
        str(param_types),
        str(class_attributes),
        pprint.pformat(kwargs) + '\n'+ str([ [k, v, type(v)] for k,v in kwargs.items()])))

    event_name = kwargs.get('event_name')
    event_params = kwargs.get('params')
    event_ctx = kwargs.get('context')
    if event_name == 'provide-client-params.s3.PutObject' and event_params:
        kensu_put(event_params=event_params, event_ctx=event_ctx, **kwargs)

#boto3._get_default_session().events.register('creating-resource-class.s3.ServiceResource',kensu_tracker)
#boto3._get_default_session().events.register('before-send.s3.PutObject', kensu_tracker)
boto3._get_default_session().events.register('provide-client-params.s3.PutObject', kensu_tracker)

# in case we wanted to see all events - use *
#S3 = boto3.resource('s3' , region_name='eu-west-1')
#event_system = S3.meta.client.meta.events
#event_system.register("*",kensu_tracker)
#event_system.register('creating-resource-class.s3.*', prt)


from abc import ABC, abstractmethod
import json
import logging
import re
import time
from kensu.utils.helpers import singleton, to_snake_case

class Reporter(object):

    def __init__(self, config):
        self.config = config

    # function taking three args: the entity, the api client, and the reporting method (e.g. `report_datastats`)
    @abstractmethod
    def apply(self, obj, kensu_api, method):
        pass 

    def entity_to_json_event(self, entity, kensu_api):
        # FIXME: check if there's a nicer way than to call kensu_api.api_client.sanitize_for_serialization
        sanitized_body = kensu_api.api_client.sanitize_for_serialization(entity)
        # p.s. serialization based on existing rules in kensu-client-scala
        now = round(1000.0 * time.time())
        offline_entity = {
            "action": "add_entity",
            "entity": re.sub(r'(.)([A-Z])', r'\1_\2', entity.__class__.__name__).upper(),
            "generatedEntityGUID": "empty",
            "schemaVersion": "0.1",
            "jsonPayload": sanitized_body,
            "context": {
                "clientId": "",
                # FIXME: make this configurable for reporting events in past
                "clientEventTimestamp": now,
                "serverReceivedTimestamp": now
            }
        }
        return json.dumps(offline_entity) + "\n"
    
    @staticmethod
    def create(config, name = None):
        name = name or config.get("name", None)
        reporter = None
        # if name == "GenericReporter":
        #     reporter = GenericReporter(config)
        if name == "DoNothingReporter":
            reporter = DoNothingReporter(config)
        elif name == "PrintReporter":
            reporter = PrintReporter(config)
        elif name == "LoggingReporter":
            reporter = LoggingReporter(config)
        elif name == "FileReporter":
            reporter = FileReporter(config)
        elif name == "ApiReporter":
            reporter = ApiReporter(config)
        elif name == "KafkaReporter":
            reporter = KafkaReporter(config)
        elif name == "MultiReporter":
            reporter = MultiReporter(config)
        return reporter


class GenericReporter(Reporter):
    def __init__(self, config, fun):
        super().__init__(config)
        self.fun = fun

    def apply(self, obj, kensu_api, method):
        return self.fun(obj, kensu_api, method)


class DoNothingReporter(Reporter):
    def apply(self, obj, kensu_api, method):
        return obj


class PrintReporter(Reporter):
    def apply(self, obj, kensu_api, method):
        json = self.entity_to_json_event(obj, kensu_api)
        print(json)
        return obj


class LoggingReporter(Reporter):
    def __init__(self, config, level=None):
        super().__init__(config)
        self.level = level
        if self.level is None and config is not None and config["level"] is not None:
            self.level = config["level"]
        self.level = self.level.lower()
        if self.level == "info":
            self.log = logging.info
        elif self.level == "warn":
            self.log = logging.warning
        elif self.level == "error":
            self.log = logging.error
        elif self.level == "debug":
            self.log = logging.debug
        else:
            print('No logging level was specified for LoggingReporter, using ERROR log level')
            self.log = logging.error

    def apply(self, obj, kensu_api, method):
        json = self.entity_to_json_event(obj, kensu_api)
        if self.log:
            res = self.log(obj)
        return obj


class FileReporter(Reporter):
    def __init__(self, config, file_name=None):
        super().__init__(config)
        if config is not None:
            file_name = file_name or config["file_name"]
        self.file = open(file_name, "a")
        import atexit
        atexit.register(self.close_gracefully)

    def close(self):
        if self.file is not None:
            self.file.close()

    def close_gracefully(self):
        try:
            self.close()
        except IOError as ex:
            print('Failed closing the output file', ex)

    def apply(self, obj, kensu_api, method):
        self.file.write(self.entity_to_json_event(obj, kensu_api))
        self.file.flush()
        return obj


class ApiReporter(Reporter):
    def apply(self, obj, kensu_api, method):
        method(obj)
        return obj


class KafkaReporter(Reporter):
    def __init__(self, config):
        super().__init__(config)
        if config is not None:
            from confluent_kafka import Producer
            import socket
            self.bootstrap_servers = ",".join(json.loads(config["bootstrap_servers"]))
            conf = {'bootstrap.servers': self.bootstrap_servers,
                    'client.id': socket.gethostname()}
            self.topic = config["topic"]
            self.producer = Producer(conf)


    def apply(self, obj, kensu_api, method):
        guid = obj.to_guid()
        token_and_entity = {
            "token": kensu_api.api_client.default_headers["X-Auth-Token"],
            "entity": self.entity_to_json_event(obj, kensu_api)
        }
        s = json.dumps(token_and_entity)
        self.producer.produce(self.topic, key=guid, value=s)
        return obj


class MultiReporter(Reporter):
    def __init__(self, config):
        super().__init__(config)
        reporters_names = json.loads(config.get("reporters"))
        self.reporters = [Reporter.create(config, name = name) for name in reporters_names]

    def apply(self, obj, kensu_api, method):
        for r in self.reporters:
            r.apply(obj, kensu_api, method)



import decimal
import logging
import datetime
import re

logger = logging.getLogger(__name__)

from kensu.google.cloud.bigquery import Client
from kensu.utils.kensu_provider import KensuProvider


def compute_bigquery_stats(table_ref=None, table=None, client=None, stats_aggs=None, input_filters=None, query=None):
    if None == query == table_ref:
        raise Exception("compute_bigquery_stats requires either a table_ref or a query")
    r = {}
    kensu = KensuProvider().instance()
    client: Client = client or kensu.data_collectors['BigQuery']
    if stats_aggs is None and table is not None:
        logger.debug('Got empty statistic listing from remote service, proceeding with fallback statistic list')
        stats_aggs = generate_fallback_stats_queries(table)

    from kensu.google.cloud.bigquery.extractor import KensuBigQuerySupport

    unnest = KensuBigQuerySupport().extract_unnest(client.get_table(table_ref))

    # TODO Add nullvalue computation for REPEATED
    schema = KensuBigQuerySupport().extract_schema_fields(client.get_table(table_ref))
    non_nullable = [k.name for k in schema if k.nullable == False]
    unsupported_stats = [k.name for k in schema if k.name.count(".")>0]

    for key in non_nullable + unsupported_stats:
        if key in stats_aggs:
            stats_aggs.pop(key)


    # "dots in schemas are not supported in BigQuery,
    # as a workaround we flatten the schema and replace the infix dots with a "ksu" string.
    # once the stats computation has done is job we convert back to the dotted notation for displaying stats"

    selector = ",".join([sql_aggregation + " " + col.replace(".","__ksu__") + "_" + stat_name
                         for col, stats_for_col in stats_aggs.items()
                         for stat_name, sql_aggregation in stats_for_col.items()])

    for nested in sorted(unnest,key=len,reverse=True):
        if "." in nested:
            selector=selector.replace(nested,nested.replace(".","__ksu__"))

    filters = ''
    if input_filters is not None and len(input_filters) > 0:
        filters = f"WHERE {' AND '.join(input_filters)}"
    if unnest != []:
        list_unnest =[]
        for el in unnest:
            list_unnest.append(f",UNNEST({el}) AS {el.replace('.','__ksu__')}")

    #This needs to be adapted i we need RECORD stats
    #i.e. with unnest = "".join(list_unnest) if unnest != [] else ''

    unnest =''
    if query:
        stats_query = f"select {selector}, sum(1) as nrows from ({query}){unnest}"
    elif table_ref:
        stats_query = f"select {selector}, sum(1) as nrows from `{str(table_ref)}`{unnest} {filters}"
    #TODO extract this and add to dim-sql and fallback stats
    stats_query = stats_query.replace("sum(case","COUNTIF(").replace("when null then 1 else 0 end)","IS NULL)").replace("when true then 1 else 0 end)","IS TRUE)")

    logger.debug(f"stats query for table {table_ref}: {stats_query}")
    from google.cloud.bigquery.job import QueryJobConfig
    job_conf = QueryJobConfig(labels = {'kensu':'kensu_stats_computation'})
    for row in client.query(stats_query, job_config = job_conf ).result():
        # total num rows (independent of column)
        #FIXME : What about this?
        if row.get('nrows'):
            r['nrows'] = row['nrows']
        # extract column specific stats
        for col, stat_names in stats_aggs.items():
            for stat_name in stat_names.keys():
                v = row[col.replace(".","__ksu__") + "_" + stat_name]
                if v.__class__ in [datetime.date, datetime.datetime, datetime.time]:
                    v = int(v.strftime("%s") + "000")
                if v.__class__ in [decimal.Decimal]:
                    v = float(v)
                if v is None:
                    v = 0
                r[(col + "." + stat_name).replace("__ksu__",".")] = v
        break  # there should be only one row here
    return r


def generate_fallback_stats_queries(table):
    stats_aggs = {}

    from kensu.google.cloud.bigquery.extractor import KensuBigQuerySupport
    schema = KensuBigQuerySupport().extract_schema_fields(table)

    for f in schema:
        # f.field_type is
        # ["STRING", "BYTES",
        # "INTEGER", "INT64",
        # "FLOAT", "FLOAT64",
        # "BOOLEAN", "BOOL",
        # "TIMESTAMP", "DATE", "TIME", "DATETIME", "GEOGRAPHY", "NUMERIC", "BIGNUMERIC",
        # "RECORD", "STRUCT",]
        # FIXME: decimal
        if f.field_type in ["INTEGER", "INT", "FLOAT", "FLOAT64", "NUMERIC", "BIGNUMERIC"]:
            stats_aggs[f.name] = {"min": f"min({f.name})",
                                  "max": f"max({f.name})",
                                  "mean": f"avg({f.name})",
                                  "nullrows": f"sum(case {f.name} when null then 1 else 0 end)"}
        elif f.field_type in ["TIMESTAMP", "DATE", "TIME", "DATETIME"]:
            stats_aggs[f.name] = {"min": f"min({f.name})",
                                  "max": f"max({f.name})",
                                  "nullrows": f"sum(case {f.name} when null then 1 else 0 end)"}
        elif f.field_type in ["BOOLEAN", "BOOL"]:
            stats_aggs[f.name] = {"true": f"sum(case {f.name} when true then 1 else 0 end)",
                                  "nullrows": f"sum(case {f.name} when null then 1 else 0 end)"}
        elif f.field_type in ["STRING"]:
            stats_aggs[f.name] = {"levels": f"count(distinct {f.name})",
                                  "nullrows": f"sum(case {f.name} when null then 1 else 0 end)"}

    return stats_aggs



import logging
import traceback
import uuid

from pglast import parse_sql, ast

from kensu.pandas import KensuPandasSupport
from kensu.utils.dsl.extractors.external_lineage_dtos import KensuDatasourceAndSchema, ExtDependencyEntry, \
    GenericComputedInMemDs
from kensu.psycopg2.pghelpers import get_table_schema, get_current_db_info


def parse_and_report(cur,
                     final_sql,
                     argslist  # type: list
                     ):
    try:
        # FIXME: bytes decoding might give some issues...
        final_sql = final_sql.decode(cur.connection.encoding)
        cur_catalog, cur_schema = get_current_db_info(cur)

        for stmt in parse_sql(final_sql):
            if isinstance(stmt, ast.RawStmt):
                stmt = stmt.stmt
                if isinstance(stmt, ast.InsertStmt):
                    parse_insert(cur, cur_catalog, cur_schema, stmt, argslist=argslist)
                elif isinstance(stmt, ast.UpdateStmt):
                    parse_update(cur, cur_catalog, cur_schema, stmt, argslist=argslist)
                else:
                    logging.info("Kensu - met a not yet supported SQL statement:")
                    logging.info("Kensu - not yet supported SQL statement:" + str(stmt) + "SQL: " + final_sql)
    except:
        logging.warning("Failed parsing a SQL statement")
        traceback.print_exc()
        traceback.print_stack()


def parse_update(cur, cur_catalog, cur_schema, stmt, argslist):
    out_table_qualified_name = format_relation_name(stmt.relation, cur_catalog=cur_catalog, cur_schema=cur_schema)
    if stmt.targetList:
        explicit_column_names = [c.name for c in stmt.targetList]
    else:
        explicit_column_names = []

    out_stats_data_pandas = None
    # explicit param_field_names of those inside argslist
    if stmt.fromClause and stmt.fromClause[0].alias and stmt.fromClause[0].alias.colnames:
        # FIXME: take union of out_columns & param_field_names
        argslist_columns = [c.val for c in stmt.fromClause[0].alias.colnames]
        #print('argslist_columns', argslist_columns)
        out_stats_data_pandas=datastats_data(argslist_columns, argslist)
        # FIXME: and here we probably need to filter only columns used in UPDATE SET clause

    # if unable to figure out, assume write columns to be all the columns
    out_columns = schema_of_used_cols_or_all(cur, stmt.relation, explicit_column_names)
    report_write((out_table_qualified_name, out_columns),
                 op_type='psycopg2 update',
                 out_stats_data_pandas=out_stats_data_pandas,
                 inputs=None)


def schema_of_used_cols_or_all(cur, relation, explicit_column_names):
    if not explicit_column_names:
        explicit_column_names = []
    orig_out_table_name = format_relation_name(relation)
    out_table_schema = get_table_schema(cur, orig_out_table_name)
    if explicit_column_names:
        out_columns = [c for c in out_table_schema
                       if c['field_name'] in explicit_column_names]
    else:
        out_columns = out_table_schema
    return out_columns


def datastats_data(columns, argslist):
    import pandas as pd_orig
    # data = [['tom', 10], ['nick', 15], ['juli', 14]]
    try:
        if len(argslist) == 0 or len(argslist[0]) != len(columns):
            logging.warning("Kensu - failed converting argslist to pandas.DataFrame: columns count do not match")
            # print("columns: {}\n arglist item: {}".format(columns, argslist[0]))
            return None
        else:
            return pd_orig.DataFrame(argslist, columns=columns)
    except:
        logging.warning("Kensu - failed converting argslist to pandas.DataFrame")
        traceback.print_stack()
        return None

def parse_insert(cur, cur_catalog, cur_schema, stmt, argslist):
    out_table_qualified_name = format_relation_name(stmt.relation, cur_catalog=cur_catalog, cur_schema=cur_schema)
    if stmt.cols:  # 'INSERT INTO table_name(column1, column2) VALUES ...'
        explicit_column_names = [c.name for c in stmt.cols]
    else: # 'INSERT INTO table_name  VALUES ...'
        explicit_column_names = []
    out_columns = schema_of_used_cols_or_all(cur, stmt.relation, explicit_column_names)
    # here out_columns are also columns of argslist
    report_write((out_table_qualified_name, out_columns),
                 op_type='psycopg2 insert',
                 out_stats_data_pandas=datastats_data(columns = [c['field_name'] for c in out_columns],
                                                      argslist=argslist),
                 inputs=None)


def format_relation_name(relation, cur_catalog=None, cur_schema=None):
    parts = [
        relation.catalogname or cur_catalog,
        relation.schemaname or cur_schema,
        relation.relname
    ]
    return '.'.join([n for n in parts if n])


def report_write(out_table, op_type, out_stats_data_pandas, inputs=None):
    # FIXME: logical name etc?
    from kensu.utils.kensu_provider import KensuProvider
    ksu = KensuProvider().instance()
    if not inputs and len(ksu.inputs_ds)>0:
        inputs=ksu.inputs_ds

    if not inputs:
        input_path = 'in-mem://'+str(uuid.uuid4())
        input_ds = KensuDatasourceAndSchema.for_path_with_opt_schema(ksu=ksu,
                                                                     ds_path=input_path,
                                                                     maybe_schema=[("unknown", "unknown")],
                                                                     ds_name=input_path,
                                                                     f_get_stats=None)
        inputs = [input_ds]

    dest_name, out_schema = out_table
    dest_path = 'postgres://{}'.format(dest_name)
    if out_schema is None:
        out_schema = [("unknown", "unknown")]
    else:
        out_schema = [(f.get('field_name') or 'unknown', f.get('field_type') or 'unknown') for f in out_schema]
    f_get_stats =lambda: None  # no stats by default
    if out_stats_data_pandas is not None:
        f_get_stats = lambda: KensuPandasSupport().extract_stats(out_stats_data_pandas)
    output_ds = KensuDatasourceAndSchema.for_path_with_opt_schema(ksu=ksu,
                                                                  ds_path=dest_path,
                                                                  maybe_schema=out_schema,
                                                                  ds_name=dest_name,
                                                                  f_get_stats=f_get_stats,
                                                                  format='postgres table')


    lineage_info = [
        ExtDependencyEntry(
        input_ds=input_ds,
            # all to all lineage
        lineage=dict([(str(out_fieldname), [in_field for in_field in input_ds.field_names()])
                      for (out_fieldname, dtype) in out_schema]))
        for input_ds in inputs
    ]
    inputs_lineage = GenericComputedInMemDs(inputs=inputs, lineage=lineage_info)
    # register lineage in KensuProvider, if any
    inputs_lineage.report(ksu=ksu,
                          df_result=output_ds,
                          operation_type=op_type,
                          report_output=True,
                          register_output_orig_data=True  # FIXME?
                          )
    if lineage_info:
        # actuly report the lineage and the write operation to the sink
        ksu.report_with_mapping()



import google
from airflow.providers.google.cloud.operators.bigquery import *

from airflow.providers.google.cloud.hooks.bigquery import BigQueryHook, BigQueryJob
from airflow.providers.google.cloud.operators import bigquery as gcp_bigquery

import logging

from google.cloud.bigquery import TableReference

from kensu.airflow.kensu_airflow_collector import COLLECTOR_STATUS_INIT, COLLECTOR_STATUS_DONE, log_status, \
    airflow_init_kensu, handle_ex
from kensu.google.cloud.bigquery import QueryJob
from typing import TYPE_CHECKING

from kensu.utils.dsl.extractors.external_lineage_dtos import GenericComputedInMemDs
from kensu.utils.helpers import extract_ksu_ds_schema
from kensu.utils.kensu_provider import KensuProvider

if TYPE_CHECKING:
    from airflow.utils.context import Context


class BigQueryInsertJobOperator(gcp_bigquery.BigQueryInsertJobOperator):
    def __init__(
            self,
            *args,
            **kwargs,
    ):
        super(BigQueryInsertJobOperator, self).__init__(
            *args, **kwargs
        )
        self.ksu_bq_job = None

    def _submit_job(
            self,
            hook: BigQueryHook,
            job_id: str,
    ) -> BigQueryJob:
        # just store bigquery job to be accessed later
        bq_job = super(BigQueryInsertJobOperator, self)._submit_job(hook, job_id)
        self.ksu_bq_job = bq_job
        return bq_job

    def execute(self, context: 'Context') -> None:
        job_id = super(BigQueryInsertJobOperator, self).execute(context)
        try:
            log_status(self, COLLECTOR_STATUS_INIT)
            airflow_init_kensu(airflow_operator=self)
            # it seems hook.get_job() shall be called only when a running non-finished job already existed (a conflict)
            bq_job = self.ksu_bq_job or self.hook.get_job(
                project_id=self.project_id,
                location=self.location,
                job_id=job_id,
            )
            bq_client = self.hook.get_client(project_id=self.project_id)
            logging.info(f"bq_job: {bq_job} type={str(type(bq_job))}")
            # Union[CopyJob, QueryJob, LoadJob, ExtractJob]
            if isinstance(bq_job, google.cloud.bigquery.job.QueryJob):
                QueryJob.report_bq_sql_query_job(
                    client=bq_client,
                    job=bq_job
                )
        except Exception as ex:
            handle_ex(self, ex)
        log_status(self, COLLECTOR_STATUS_DONE)
        return job_id


BigQueryInsertJobOperator.__doc__ = gcp_bigquery.BigQueryInsertJobOperator.__doc__


class BigQueryCreateExternalTableOperator(gcp_bigquery.BigQueryCreateExternalTableOperator):

    def __init__(
            self,
            *args,
            **kwargs,
    ):
        super(BigQueryCreateExternalTableOperator, self).__init__(
            *args, **kwargs
        )

    def execute(self, context: 'Context') -> None:
        res = super(BigQueryCreateExternalTableOperator, self).execute(context)
        try:
            log_status(self, COLLECTOR_STATUS_INIT)
            # logging.info("Airflow context keys: " + str(context.keys()))
            # logging.info("Airflow context: " + str(context.items()))
            airflow_init_kensu(airflow_operator=self)
            ksu = KensuProvider().instance()
            if self.table_resource:
                bq_hook = BigQueryHook(
                    gcp_conn_id=self.bigquery_conn_id,
                    delegate_to=self.delegate_to,
                    location=self.location,
                    impersonation_chain=self.impersonation_chain,
                )
                #  e.g. table_resource: {
                #             "externalDataConfiguration": {
                #                 "sourceFormat": "PARQUET",
                #                 "sourceUris": [f"gs://{BUCKET}/raw/{parquet_file}"],
                #             },
                # }
                ext_input_conf = self.table_resource.get('externalDataConfiguration', {})
                source_uris = [f"gs://{self.bucket}/{source_object}"
                               for source_object in (self.source_objects or [])] or \
                              ext_input_conf.get('sourceUris')
                source_format = ext_input_conf.get('sourceFormat') or 'CSV'

                bq_table_ref_dict = self.table_resource['tableReference']
                bq_client = bq_hook.get_client(project_id=bq_table_ref_dict['projectId'])

                bq_table_ref = TableReference.from_api_repr(bq_table_ref_dict)
                dst_bq_table = bq_client.get_table(bq_table_ref)

                _, result_schema = extract_ksu_ds_schema(ksu, dst_bq_table)
                ksu_inputs, lineage = GenericComputedInMemDs.build_approx_lineage(
                    source_uris=source_uris,
                    source_format=source_format,
                    result_schema=result_schema
                )
                if not QueryJob.report_ddl_write_with_stats(result=dst_bq_table,
                                                            ddl_target_table=dst_bq_table,
                                                            client=bq_client,
                                                            lineage=lineage,
                                                            is_ddl_write=True,
                                                            operation_type='BigQueryCreateExternalTableOperator'):
                    logging.info("inputs:" + str(ksu_inputs) + f"self.source_objects: {self.source_objects}")
            else:
                logging.warning("table_resource not used - not supported by Kensu")
        except Exception as ex:
            handle_ex(self, ex)
        log_status(self, COLLECTOR_STATUS_DONE)
        return res


BigQueryCreateExternalTableOperator.__doc__ = gcp_bigquery.BigQueryCreateExternalTableOperator.__doc__



# coding: utf-8

"""
    

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)

    OpenAPI spec version: beta
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


from __future__ import absolute_import

# import ApiClient
from .api_client import ApiClient
from .apis.kensu_entities_api import KensuEntitiesApi
# import apis into sdk package
from .configuration import Configuration
# import models into sdk package
from .models.batch_code_base import BatchCodeBase
from .models.batch_code_version import BatchCodeVersion
from .models.batch_data_source import BatchDataSource
from .models.batch_data_stats import BatchDataStats
from .models.batch_entity_report import BatchEntityReport
from .models.batch_entity_report_result import BatchEntityReportResult
from .models.batch_error_message import BatchErrorMessage
from .models.batch_lineage_run import BatchLineageRun
from .models.batch_model import BatchModel
from .models.batch_model_metrics import BatchModelMetrics
from .models.batch_model_training import BatchModelTraining
from .models.batch_physical_location import BatchPhysicalLocation
from .models.batch_process import BatchProcess
from .models.batch_process_lineage import BatchProcessLineage
from .models.batch_process_run import BatchProcessRun
from .models.batch_process_run_stats import BatchProcessRunStats
from .models.batch_project import BatchProject
from .models.batch_schema import BatchSchema
from .models.batch_schema_field_tag import BatchSchemaFieldTag
from .models.batch_user import BatchUser
from .models.code_base import CodeBase
from .models.code_base_pk import CodeBasePK
from .models.code_base_ref import CodeBaseRef
from .models.code_version import CodeVersion
from .models.code_version_pk import CodeVersionPK
from .models.code_version_ref import CodeVersionRef
from .models.data_source import DataSource
from .models.data_source_pk import DataSourcePK
from .models.data_source_ref import DataSourceRef
from .models.data_stats import DataStats
from .models.data_stats_pk import DataStatsPK
from .models.field_def import FieldDef
from .models.lineage_run import LineageRun
from .models.lineage_run_pk import LineageRunPK
from .models.lineage_run_ref import LineageRunRef
from .models.model import Model
from .models.model_metrics import ModelMetrics
from .models.model_metrics_pk import ModelMetricsPK
from .models.model_pk import ModelPK
from .models.model_ref import ModelRef
from .models.model_training import ModelTraining
from .models.model_training_pk import ModelTrainingPK
from .models.model_training_ref import ModelTrainingRef
from .models.physical_location import PhysicalLocation
from .models.physical_location_pk import PhysicalLocationPK
from .models.physical_location_ref import PhysicalLocationRef
from .models.process import Process
from .models.process_lineage import ProcessLineage
from .models.process_lineage_pk import ProcessLineagePK
from .models.process_lineage_ref import ProcessLineageRef
from .models.process_pk import ProcessPK
from .models.process_ref import ProcessRef
from .models.process_run import ProcessRun
from .models.process_run_pk import ProcessRunPK
from .models.process_run_ref import ProcessRunRef
from .models.process_run_stats import ProcessRunStats
from .models.process_run_stats_pk import ProcessRunStatsPK
from .models.project import Project
from .models.project_pk import ProjectPK
from .models.project_ref import ProjectRef
from .models.schema import Schema
from .models.schema_lineage_dependency_def import SchemaLineageDependencyDef
from .models.schema_pk import SchemaPK
from .models.schema_ref import SchemaRef
from .models.user import User
from .models.user_pk import UserPK
from .models.user_ref import UserRef

configuration = Configuration()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#
# Copyright 2021 Kensu Inc
#
import logging
import traceback

import google

from kensu.google.cloud.bigquery.extractor import KensuBigQuerySupport
from kensu.google.cloud.bigquery.job.bigquery_stats import compute_bigquery_stats
from kensu.google.cloud.bigquery.job.offline_parser import BqOfflineParser
from kensu.google.cloud.bigquery.job.remote_parser import BqRemoteParser
from kensu.pandas import DataFrame
from kensu.utils.helpers import report_all2all_lineage
from kensu.utils.kensu_provider import KensuProvider
import google.cloud.bigquery as bq
import google.cloud.bigquery.job as bqj

logger = logging.getLogger(__name__)


class QueryJob(bqj.QueryJob):
    # - client.query(sql) returns QueryJob which we patch here for tracking
    # - for both QueryJob.to_dataframe and QueryJob.result() we return a monkey-patched QueryJob
    #      having QueryJob.result set to our wrapper fn
    #  * we add a .ksu_dest field to QueryJob.result() which is internally used to connect lineage later on
    # - QueryJob.to_dataframe calls original .result() internally, so to be able to track it, currently
    # we do not call the original .to_dataframe(), but obtain .result() and then convert it to DataFrame
    # Limitations:
    # - currently bq.TableReference (temp table) is tested as destination of job (could be RowIterator too)

    @staticmethod
    def patch(job: bqj.QueryJob) -> bqj.QueryJob:
        return QueryJob.override_result(QueryJob.override_to_dataframe(job))

    @staticmethod
    def override_to_dataframe(job: bqj.QueryJob) -> bqj.QueryJob:
        f_orig_to_dataframe = job.to_dataframe

        def wrapper(*args, **kwargs):
            # instead of calling .to_dataframe(),  to be able to track lineage
            # we get raw .result() and later convert result to pandas
            # (otherwise original input info is lost inside to_dataframe fn)
            job_result = job.result()
            to_dataframe_res = job_result.to_dataframe(*args, **kwargs)
            final_result = DataFrame.using(to_dataframe_res)
            report_all2all_lineage(in_obj=job_result, out_obj=to_dataframe_res, in_inmem=True, out_inmem=True,
                                   op_type='to_dataframe')
            return final_result

        wrapper.__doc__ = f_orig_to_dataframe.__doc__
        setattr(job, 'to_dataframe', wrapper)
        return job

    @staticmethod
    def override_result(job  # type: google.cloud.bigquery.job.QueryJob
                        ) -> bqj.QueryJob:
        f = job.result

        def wrapper(*args, **kwargs):
            kensu = KensuProvider().instance()
            result = f(*args, **kwargs)
            client = kensu.data_collectors['BigQuery']
            QueryJob.report_bq_sql_query_job(client=client, job=job, result=result)
            return result

        wrapper.__doc__ = f.__doc__
        setattr(job, 'result', wrapper)
        return job


    @staticmethod
    def report_bq_sql_query_job(
            client: 'Client',
            job: 'QueryJob',
            result=None
    ):
        """track QueryJob without requiring to subclass the Job itself"""

        kensu = KensuProvider().instance()
        ddl_target_table = job.ddl_target_table
        dest = job.destination or ddl_target_table
        is_ddl_write = bool(ddl_target_table)
        logger.debug(f'in QueryJob.result(): dest={dest}')
        logger.debug(f'in QueryJob.result(): referenced={job.referenced_tables}')

        if isinstance(dest, bq.TableReference):
            dest = client.get_table(dest)

        if result is not None:
            QueryJob._store_bigquery_job_destination(result=result, dest=dest)
        else:
            if isinstance(dest, bq.Table):
                result = dest
            else:
                return


        # TODO What if several SELECT queries linked with ; ?
        db_metadata, table_id_to_bqtable, table_infos = BqOfflineParser.get_referenced_tables_metadata(
            kensu=kensu,
            client=client,
            job=job,
            query=BqOfflineParser.normalize_table_refs(job.query))
        try:
            bq_lineage = BqRemoteParser.parse(
                kensu=kensu,
                client=client,
                query=job.query,
                db_metadata=db_metadata,
                table_id_to_bqtable=table_id_to_bqtable)
        except:
            logger.warning("Error in BigQuery collector, using fallback implementation")
            traceback.print_exc()
            bq_lineage = BqOfflineParser.fallback_lineage(kensu, table_infos, dest)
        QueryJob.report_ddl_write_with_stats(
            result=result,
            ddl_target_table=ddl_target_table,
            client=client,
            lineage=bq_lineage,
            is_ddl_write=is_ddl_write
        )


    @staticmethod
    def report_ddl_write_with_stats(
            result,
            ddl_target_table,
            client,
            lineage,
            is_ddl_write,
            operation_type=None
    ):
        kensu = KensuProvider().instance()
        if kensu.compute_stats and is_ddl_write and isinstance(ddl_target_table, bq.Table):
            # for DDL writes, stats can be computed by just reading the whole table
            # FIXME: for incremental `INSERT INTO` would not give the correct stats (we'd get full table)
            out_stats_values = compute_bigquery_stats(
                table_ref=ddl_target_table.reference,
                table=ddl_target_table,
                client=client,
                # stats for all output columns
                stats_aggs=None,
                input_filters=None)
            KensuBigQuerySupport().set_stats(result, out_stats_values)
        lineage.report(
            ksu=kensu,
            df_result=result,
            operation_type=operation_type,
            report_output=kensu.report_in_mem or is_ddl_write,
            register_output_orig_data=is_ddl_write   # used for DDL writes, like CREATE TABLE t2 AS SELECT * FROM t1
        )
        if is_ddl_write:
            if len(lineage.lineage) > 0:
                kensu.report_with_mapping()
                return True
            else:
                logger.warning("Kensu got empty lineage - not reporting")
                logger.info("lineage:" + str(lineage))
                logger.info("ddl_target_table:" + str(ddl_target_table))
                return False


    @staticmethod
    def _store_bigquery_job_destination(result, dest):
        try:
            if (dest is not None) and isinstance(dest, google.cloud.bigquery.table.Table):
                result.ksu_dest = dest
        except:
            logger.warning('setting _store_bigquery_job_destination failed')
            traceback.print_exc()



import json as js
from json import *

from kensu.boto3 import ksu_dict
from kensu.botocore.response import ksu_bytes
from kensu.client import DataSourcePK, DataSource, FieldDef, SchemaPK, Schema
from kensu.itertools import kensu_list
from kensu.utils.dsl.extractors.external_lineage_dtos import KensuDatasourceAndSchema
from kensu.utils.helpers import logical_naming_batch, extract_short_json_schema
from kensu.utils.kensu_provider import KensuProvider


def wrap_loads(method):
    def wrapper(*args, **kwargs):

        result = method(*args, **kwargs)

        kensu = KensuProvider().instance()

        data = args[0]
        if isinstance(result,dict):
            result = ksu_dict(result)
        elif isinstance(result,list):
            result = kensu_list(result)
        if isinstance(data,ksu_bytes):
            result.ksu_metadata = data.ksu_metadata

            origin_location = result.ksu_metadata['origin_location']
            origin_name = result.ksu_metadata['origin_name']

            if kensu.logical_naming == 'ReplaceNumbers':
                logical = logical_naming_batch(origin_name)
            else:
                logical = origin_name

            #FIXME we should put this in extractors
            result_pk = DataSourcePK(location=origin_location,
                                     physical_location_ref=kensu.default_physical_location_ref)
            result_ds = DataSource(name=origin_name, categories=['logical::' + logical], format=origin_name.split('.')[-1],
                                   pk=result_pk)._report()

            short_result_sc = extract_short_json_schema(result, result_ds)._report()

            kensu.real_schema_df[short_result_sc.to_guid()] = None

            kensu.add_input_ref(KensuDatasourceAndSchema(ksu_ds=result_ds, ksu_schema=short_result_sc))

        return result

    wrapper.__doc__ = method.__doc__
    return wrapper

loads = wrap_loads(js.loads)


import os
# disable Kensu collector when requested to do so (e.g. inside Apache Spark executor nodes)
if "KSU_DISABLE_PY_COLLECTOR" in os.environ:
    from sklearn import model_selection
    from sklearn.linear_model import LogisticRegression
else:
    from kensu.sklearn import model_selection
    from kensu.sklearn.linear_model import LogisticRegression



from hashlib import sha1

class SimpleCache(object):
  def __init__(self):
    self.CACHE = {}
  def hash(self, o):
    return sha1(str(o).encode()).hexdigest()
  def put(self, k, o):
    kh = self.hash(k)
    self.CACHE[kh] = o
  def contains(self, o):
    k = self.hash(o)
    return k in self.CACHE
  def get(self, o):
    if self.contains(o):
      k = self.hash(o)
      return self.CACHE[k]
    else:
      return None